/*
 This program is free software: you can use, modify and/or
 redistribute it under the terms of the simplified BSD
 License. You should have received a copy of this license along
 this program. If not, see
 <http://www.opensource.org/licenses/bsd-license.html>.
*/

#include <Image.h>

/**
 * @file    Image.cpp
 * @brief   Functions of the class 
 *          Image for accessing to its data 
 *          and to modify it.
 *
 * @author  Sira Ferradans
 */

/**
 * @function    Contructors of the class Image
 * @brief   Functions for creating emty objects
 *          of the class Image
 *
 */
Image::Image()
{
  dim[0]=0;
  dim[1]=0;
  data=new double[1];
}

/**
 * @function    Contructors of the class Image
 * @brief       Constructor of the class Image 
 *              that allocate memory for the selected dimensions
 *
 * @param      row,col: size of the image
 *
 *
 **/
Image::Image(int row, int col)
{
  dim[0]=row;
  dim[1]=col;
  data=new double[row*col];
}

/**
 * @function    Contructors of the class Image
 * @brief       Constructor of the class Image 
 *              that sets all values to 'val'
 *
 * @param      row,col: size of the image
 *
 * @param      val: all elements in the image
 *                  are set to this value
 *
 **/
Image::Image(int row, int col, double val)
{
// create image with values set to 'val'
  dim[0]=row;
  dim[1]=col;
  data=new double[row*col];
    
  for(int i=0; i<row*col;i++)
    data[i]=val;
}

/**
 * @function    Contructors of the class Image
 * @brief       Image copy constructor. Construct a new image
 *              and copy the existing data
 *
 * @param      im2:     image to be copied
 *
 **/
Image::Image(Image  & im2)
{
    dim[0]=im2.rows();
    dim[1]=im2.cols();
    
    data=new double[dim[0]*dim[1]];
    
    int length=dim[0]*dim[1];
    for(int i=0; i<length; i++)
        data[i]=im2.data[i];
    
    return;
}

/**
 * @function    Destuctor of the class Image
 * @brief   Functions for destroying images
 *
 */
Image::~Image()
{
  delete[] data;
}

/**
 * @function    Operators of the class Image
 * @brief       Operators of the class Image that 
                modify its values 'point by point' 
                multiplying each for a constant
 *
 * @param      dt: constant to multiply over all image values
 *
 **/
void Image::operator*=(double dt)
{
  int length=dim[0]*dim[1];
    
  for(int i=0;i<length;i++)
    data[i]*=dt;
}

/**
 * @function    Operators of the class Image
 * @brief       Operators of the class Image that modify 
 *              its values 'point by point' multiplying
 *              each value between images
 *
 * @param      im2: image to multiply with the original one
 *
 **/
void Image::operator*=(Image  & im2)
{
  int row2=im2.rows();
  int col2=im2.cols();
    
  if(dim[0]!=row2 || dim[1]!=col2)
    {
      fprintf(stderr,"Error: Image::Operator*=: Cannot multiply");
      fprintf(stderr,"because we have different dimensions !\n");
      return;
    }

  int length=dim[0]*dim[1];
    
  for(int i=0; i<length; i++)
    this->data[i] *= (&im2)->data[i];

  return;
}

/**
 * @function    Operators of the class Image
 * @brief       Copy operator, which copy each image value
 *
 * @param      im2: image to be copied into the original one
 *
 **/
void Image::operator=(Image  & im2)
{
  int row2=im2.rows();
  int col2=im2.cols();
    
  if(dim[0]!=row2 || dim[1]!=col2)
    {
      dim[0]=row2;
      dim[1]=col2;
      
      delete[] data;
      data=new double[row2*col2];
    }

  int length=dim[0]*dim[1];
  
  for(int i=0; i<length; i++)
    data[i]=im2.data[i];
}

/**
 * @function    Operators of the class Image
 * @brief       Addition operation between images
 *
 * @param      im2: image to be added to the original one
 *
 **/
void Image::operator+=(Image & im2)
{
    int row2=im2.rows();
    int col2=im2.cols();
    
    if(dim[0]!=row2 || dim[1]!=col2)
    {
        fprintf(stderr,"Warning: different dim. to sum \n");
        return;
    }
    
    int length=dim[0]*dim[1];
    
    for(int i=0; i<length; i++)
        data[i]+=im2.data[i];
}

/**
 * @function    Operators of the class Image
 * @brief       Addition operation between image and constant,
 *              which add the constar to each value of the image
 *
 * @param      dt: constant to be added to each image point
 *
 **/
void Image::operator+=(double dt)
{
    int length=dim[0]*dim[1];
    for(int i=0;i<length;i++)
        data[i]+=dt;
}

/**
 * @brief       Function to allocate memrory for an existing image
 *
 * @param      row,col: size of the image
 *
 **/
void Image::initialize(int row, int col)
{
    this->dim[0]=row;
    this->dim[1]=col;
    
    delete[] this->data;
    this->data=new double[row*col];
}

/**
 * @brief       Finding the min value of the image
 *
 * @return min value of the image
 **/
double Image::minval()
{
  double m=1e20;
  int length=dim[0]*dim[1];
    
  for(int i=0; i<length; i++)
  {
    if(data[i] < m)
      m = data[i];
  }
    
  return m;
}

/**
 * @brief       Finding the max value of the image
 *
 * @return max value of the image
 **/
double Image::maxval()
{
  double m=-1e20;
  int length=dim[0]*dim[1];
    
  for(int i=0; i<length; i++)
  {
      if(data[i] > m)
      {
          m = data[i];
      }
  }

  return m;
}

/**
 * @brief       Finding the max absolut value of the image
 *
 * @return max absolut value of the image
 **/
double Image::maxabsval()
{
    double m = 0;
    
    int length=dim[0]*dim[1];
    
    for(int i=0; i<length; i++)
        if(fabs(data[i]) > m)
            m = fabs(data[i]);
    
    
    return m;
}

/**
 * @brief       Calculation the mean value of the image
 *
 * @return mean value of the image
 **/
double Image::meanvalue()
{
    // finds max abs value
    double m = 0;
    
    int length=dim[0]*dim[1];
    
    for(int i=0; i<length; i++)
    {
        m += data[i];
    }
    
    return m/length;
}

/**
 * @brief   Computes the squared distance between (*this) and Im2
 *
 * @param Im2: Image to be compared with the original one.
 *
 * @return Squared distance between images
 **/
double Image::Dist(Image & Im2)
{
    int length=dim[0]*dim[1];
    double res=0.0;
    double tmp;
    double v=1.0;
    double v2=1.0;
    
    for(int i=0;i<length;i++)
    {
        tmp=(data[i])*v-(Im2.data[i])*v2;
        tmp=fabs(tmp);
        res+=tmp;
    }
    return(res/length);
    
}

/**
 * @brief   Rescale the image to obtain an image in the selected range
 *
 * @param minv,maxv: minimal and maximal 
 * references values to rescale the image
 *
 **/
void Image::rescale(double minv, double maxv)
{
    int length=dim[0]*dim[1];
    
    double M=this->maxval();
    double m=this->minval();
    
    double R;
    double eps=1e-6;
    double s=(maxv-minv)/(M-m);
    for(int i=0;i<length;i++)
    {
        R=data[i];
        data[i]=minv+s*(R-m);
    }
}

/**
 * @brief   This function shifts the image swaping oposite image quarters
 *
 **/
void Image::fftshift()
{
    int row=dim[0];
    int col=dim[1];
    
    double tmp;
    
    for(int i=0;i<row/2;i++)
    {
        for(int j=0;j<col/2;j++)
        {
            tmp=(*this)(i,j);
            (*this)(i,j)=(*this)(i+row/2,j+col/2);
            (*this)(i+row/2,j+col/2)=tmp;
            tmp=(*this)(i+row/2,j);
            (*this)(i+row/2,j)=(*this)(i,j+col/2);
            (*this)(i,j+col/2)=tmp;
        }
    }
}

/**
 * @brief   It generates a gaussian kernel
 *
 * @param row,col: kernel dimensions
 *
 * @param sigma: std of the gaussian function
 *
 **/
void  Image::gaussian_kernel(int row, int col, double sigma)
{
    dim[0]=row;
    dim[1]=col;
    
    delete[] data;
    data=new double[row*col];
  
    double normaliza=1.0;

    int mitrow=row/2;
    int mitcol=col/2;
    
    for(int i=0;i<row;i++)
        for(int j=0;j<col;j++){
            (*this)(i,j)=normaliza*exp( -((i-mitrow)*(i-mitrow)+
                                          (j-mitcol)*(j-mitcol) )
                                       /(2.0*sigma*sigma));
            
        }
    
}

/**
 * @brief   It generates a linear kernel
 *
 * @param alpha: decresing velocity in the way 1/(alpha + distance)
 *
 **/
void Image::linear_kernel(double alpha)
{
    double mitrow=((double)rows())/2.0;
    double mitcol=((double)cols())/2.0;
    
    double c1 = mitrow*mitrow;
    double c2 = mitcol*mitcol;
    double d;
    
    for(int i=0;i<rows();i++)
    {
        for(int j=0;j<cols();j++)
        {
            d = sqrt( (i-mitrow)*(i-mitrow)/c1+
                     (j-mitcol)*(j-mitcol)/c2);
            (*this)(i,j)=1.0-alpha*d/(2.0*sqrt(2));
        }
    }
}