/*
 This program is free software: you can use, modify and/or
 redistribute it under the terms of the simplified BSD
 License. You should have received a copy of this license along
 this program. If not, see
 <http://www.opensource.org/licenses/bsd-license.html>.
  */

#include <contrast_enhancement.h>
#include <inout.h>

/**
 * @file    contrast_enhancement.cpp
 * @brief   Palma-Amestoy et al. method for 
 * contrast enhancement and color cast removal.
 *
 * This file implements the contrast enhancement and 
 * color cast removal algorithm described in
 *
 * “A perceptually inspired variational framework for color enhancement”
 * by R. Palma-Amestoy, E. Provenzi, M. Bertalmío,V. Caselles.IEEE Trans.
 * on Pattern Analysis and Machine Intelligence (TPAMI), 2009.
 *
 * @author  Sira Ferradans and Rodrigo Palma-Amestoy
 */

/**
 * @brief Multiplication element per element of two images with complex values
 *
 * @param A: 
 *      Left side argument.
 *
 * @param B:
 *      Rigth side argument.
 *
 * @param C: 
 *      Matrix used to save the multiplication's result.
 */
void fftproduct(fftw_complex * A, fftw_complex * B,
                fftw_complex * C, int row, int col)
{
    int length=row*(col/2.0+1.0);
    
    for (int i = 0; i < length; i++)
    {
        C[i][0] = A[i][0] * B[i][0] - A[i][1] * B[i][1];
        C[i][1] = A[i][0] * B[i][1] + A[i][1] * B[i][0];
    }
}

/**
 * @brief Evalution of the first part of the 
 * contrast enhancement polinomial aproximation.
 *
 * @param poly
 *      Polynomial coefficient array.
 * @param p
 *      Intermediate estimated polynomial coefficient.
 * @param u1...u9
 *      Image to be enhanced and its powers from 1 .. 9
 */
void transform_polynom_coef(double* poly, double *p,
                            double u1,double u2,double u3,
                            double u4,double u5,double u6,
                            double u7,double u8,double u9)
{
    p[0] = poly[0] + poly[1] * u1 +
    poly[3] * u2 +
    poly[6] * u3 +
    poly[10]* u4 +
    poly[15]* u5 +
    poly[21]* u6 +
    poly[28]* u7 +
    poly[36]* u8 +
    poly[45]* u9;
    
    p[1] = poly[2] + poly[4] * u1 +
    poly[7] * u2 +
    poly[11] * u3 +
    poly[16] * u4 +
    
    poly[22] * u5 +
    poly[29] * u6 +
    poly[37] * u7 +
    poly[46] * u8;
    
    p[2] = poly[5] + poly[8] * u1 +
    poly[12] * u2 +
    poly[17] * u3 +
    poly[23] * u4 +
    poly[30] * u5 +
    poly[38] * u6 +
    poly[47] * u7;
    
    p[3] = poly[9] + poly[13] * u1 +
    poly[18] * u2 +
    poly[24] * u3 +
    poly[31] * u4 +
    poly[39] * u5 +
    poly[48] * u6;
    
    p[4] = poly[14] + poly[19] * u1 +
                      poly[25] * u2 +
                      poly[32] * u3 +
                      poly[40] * u4 +
                      poly[49] * u5;
    p[5] = poly[20] + poly[26] * u1 +
                      poly[33] * u2 +
                      poly[41] * u3 +
                      poly[50] * u4;
    
    p[6] = poly[27] + poly[34] * u1 + poly[42] * u2 + poly[51] * u3;
    p[7] = poly[35] + poly[43] * u1 + poly[52] * u2;
    p[8] = poly[44] + poly[53] * u1;
    p[9] = poly[54];
}

/**
 * @brief Implements the contrast_enhancement algorithm for color images
 *
 * @param Im0
 *      Original image to be enhanced.
 * @param file_out
 *      Location where the result is to be saved.
 * @param poly
 *      Polynomial coefficient for the contrast function aproximation.
 * @param sigma
 *      Control on the amount of local contrast,
 *       common valid values are (1,100).
 *      The bigger the value the higher the amount 
 *      of contrast enhancement.
 * @param alpha
 *      Level of attachment to grey value.
 * @param verbose
 *      Enable/Disable verbose mode.
 *
 */
void contrast_enhancement(Image* Im0,double *poly, int type_kernel,
                          double sigma, double alpha,
                          double beta, int verbose)
{
    /* -----------------------------------------------------------------*/
    /* Starting algorithm                                                 */
    /* -----------------------------------------------------------------*/

      // get the dimensions of the data
    int col=Im0->cols(), row=Im0->rows();
    int length=row*col;
    
    // methods params
    double dt=0.2; // time step or velocity of the process
    
    // difference image Ik+1 - Ik threhold to finalize the process.
    double threshold_diff=0.001;
    // gray world (1/2 image) term's weight: control alpha>1
    alpha = max(alpha,255.0/253.0);
    // original image (I0) term's weigth, beta >0
    beta = max(beta,0);

    /* -----------------------------------------------------------------*/
    /* Gaussian kernel definition                                       */
    /* -----------------------------------------------------------------*/
    // kernel definition
    Image g(row,col,0.0);
    
    //Note that the smaller values of 'sigma'
    //imply less locality in the contrast enhancement
    if (type_kernel == 1)
        g.linear_kernel(sigma); // linear kernel 1.0/(sigma + distance)
    else
        g.gaussian_kernel(row,col,(double)min(row,col)/sigma);
        
    if (verbose){
        printf("Algorithm parameters:");
        printf("(alpha,beta,kernel type, kernel sigma)=");
        
        if (type_kernel == 1)
             printf("(%2.2f,%2.2f,linear, %2.2f)\n",alpha,(double)beta,(double)sigma);
        else
             printf("(%2.2f,%2.2f,Gaussian, %2.2f)\n",alpha,(double)beta,(double)min(row,col)/sigma);
        
        printf("saving kernel into ./kernel.tiff\n");
        writeBW("./kernel.tiff",&g); // writing image to the disc
    }
    g.fftshift();
   
    double suma = 0.0;
    for(int i=0;i<length;i++)
        suma+=g.data[i];
    g*=(1.0/suma);
    
    // we need the FFT of the kernel, and its fftw plan
    fftw_complex* G = (fftw_complex*)
                      fftw_malloc(sizeof(fftw_complex) * length);
    fftw_plan pG = fftw_plan_dft_r2c_2d(row, col, g.data, G,FFTW_ESTIMATE);
    fftw_execute(pG);
    
    /* -----------------------------------------------------------------*/
    /* Starting core process                                            */
    /* -----------------------------------------------------------------*/
    
    // auxiliar images definition
    Image u1(row,col);
    Image u2(row,col);
    Image u3(row,col);
    Image u4(row,col);
    Image u5(row,col);
    Image u6(row,col);
    Image u7(row,col);
    Image u8(row,col);
    Image u9(row,col);
    
    Image iu1(row,col);
    Image iu2(row,col);
    Image iu3(row,col);
    Image iu4(row,col);
    Image iu5(row,col);
    Image iu6(row,col);
    Image iu7(row,col);
    Image iu8(row,col);
    Image iu9(row,col);
    
    // auxiliar fft variables definition
    fftw_complex* U1;
    U1 = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*row*col);
    fftw_complex* U2;
    U2 = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*row*col);
    fftw_complex* U3;
    U3 = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*row*col);
    fftw_complex* U4 ;
    U4 = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*row*col);
    fftw_complex* U5;
    U5 = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*row*col);
    fftw_complex* U6;
    U6 = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*row*col);
    fftw_complex* U7;
    U7 = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*row*col);
    fftw_complex* U8;
    U8 = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*row*col);
    fftw_complex* U9;
    U9 = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*row*col);
    
    
    int aux_length = row*col;
    fftw_complex* U1G;
    U1G = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*aux_length);
    fftw_complex* U2G;
    U2G = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*aux_length);
    fftw_complex* U3G;
    U3G = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*aux_length);
    fftw_complex* U4G;
    U4G = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*aux_length);
    fftw_complex* U5G;
    U5G = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*aux_length);
    fftw_complex* U6G;
    U6G = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*aux_length);
    fftw_complex* U7G;
    U7G = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*aux_length);
    fftw_complex* U8G;
    U8G = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*aux_length);
    fftw_complex* U9G;
    U9G = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*aux_length);
    
    fftw_plan pU1 = fftw_plan_dft_r2c_2d(row, col, u1.data, U1,FFTW_ESTIMATE);
    fftw_plan pU2 = fftw_plan_dft_r2c_2d(row, col, u2.data, U2,FFTW_ESTIMATE);
    fftw_plan pU3 = fftw_plan_dft_r2c_2d(row, col, u3.data, U3,FFTW_ESTIMATE);
    fftw_plan pU4 = fftw_plan_dft_r2c_2d(row, col, u4.data, U4,FFTW_ESTIMATE);
    fftw_plan pU5 = fftw_plan_dft_r2c_2d(row, col, u5.data, U5,FFTW_ESTIMATE);
    fftw_plan pU6 = fftw_plan_dft_r2c_2d(row, col, u6.data, U6,FFTW_ESTIMATE);
    fftw_plan pU7 = fftw_plan_dft_r2c_2d(row, col, u7.data, U7,FFTW_ESTIMATE);
    fftw_plan pU8 = fftw_plan_dft_r2c_2d(row, col, u8.data, U8,FFTW_ESTIMATE);
    fftw_plan pU9 = fftw_plan_dft_r2c_2d(row, col, u9.data, U9,FFTW_ESTIMATE);
    
    fftw_plan pinvU1 = fftw_plan_dft_c2r_2d(row, col,
                                            U1G,iu1.data, FFTW_ESTIMATE);
    fftw_plan pinvU2 = fftw_plan_dft_c2r_2d(row, col,
                                            U2G,iu2.data, FFTW_ESTIMATE);
    fftw_plan pinvU3 = fftw_plan_dft_c2r_2d(row, col,
                                            U3G,iu3.data, FFTW_ESTIMATE);
    fftw_plan pinvU4 = fftw_plan_dft_c2r_2d(row, col,
                                            U4G,iu4.data, FFTW_ESTIMATE);
    fftw_plan pinvU5 = fftw_plan_dft_c2r_2d(row, col,
                                            U5G,iu5.data, FFTW_ESTIMATE);
    fftw_plan pinvU6 = fftw_plan_dft_c2r_2d(row, col,
                                            U6G,iu6.data, FFTW_ESTIMATE);
    fftw_plan pinvU7 = fftw_plan_dft_c2r_2d(row, col,
                                            U7G,iu7.data, FFTW_ESTIMATE);
    fftw_plan pinvU8 = fftw_plan_dft_c2r_2d(row, col,
                                            U8G,iu8.data, FFTW_ESTIMATE);
    fftw_plan pinvU9 = fftw_plan_dft_c2r_2d(row, col,
                                            U9G,iu9.data, FFTW_ESTIMATE);
    // array to save intermediate polynomial result
    double* p = (double*)malloc(10*sizeof(double));
    

  /* -----------------------------------------------------------------*/
   /* Start the gradient descent                                       */
   /* -----------------------------------------------------------------*/
   // variables to halt gradient descent
   int iteration=0;
   double difference=1000.0;
    
    // variables needed
    double mu = 0.5; // grey world assumption
    // normalization factor in the gradient descent
    double norm = 1.0/(1.0 + dt*(alpha + beta));
    
    // saving the previous step to calculate de differences Ik+1 - Ik.
    Image Im_k;
    Image R(row,col); // this image will be used to save R contrast term.
    Image Im_kp1;
    
    if(verbose)
    {
        fprintf(stderr,"Starting gradient descent\n");
    }
    
   
    for (int color =0;color <3; color++)
    {
        if(verbose)
        {
            fprintf(stderr,"Color channel:%d\n\n\n",color);
        }
        
        
        Im_kp1 = Im0[color];
         // optional image rescaling to increase the algorith capacity
        Im_kp1.rescale(0.0, 1.0);
        
        
        // main loop of gradient descent process
        while((iteration < 100) && (difference > threshold_diff))
        {
            
            /* -------------------------------------*/
            /* Convolutions  : compute w * I^j      */
            /* -------------------------------------*/
            // Ik image power 1..9 calculation
            u1=Im_kp1;
            u2=u1;
            u2*=u1;
            u3=u2;
            u3*=u1;
            u4=u3;
            u4*=u1;
            u5=u4;
            u5*=u1;
            u6=u5;
            u6*=u1;
            u7=u6;
            u7*=u1;
            u8=u7;
            u8*=u1;
            u9=u8;
            u9*=u1;
            
            // fft of Ik image power 1...9
            fftw_execute(pU1);
            fftw_execute(pU2);
            fftw_execute(pU3);
            fftw_execute(pU4);
            fftw_execute(pU5);
            fftw_execute(pU6);
            fftw_execute(pU7);
            fftw_execute(pU8);
            fftw_execute(pU9);
            
            // product between images and kernel's fft (convolution)
            fftproduct(U1,G,U1G,row,col);
            fftproduct(U2,G,U2G,row,col);
            fftproduct(U3,G,U3G,row,col);
            fftproduct(U4,G,U4G,row,col);
            fftproduct(U5,G,U5G,row,col);
            fftproduct(U6,G,U6G,row,col);
            fftproduct(U7,G,U7G,row,col);
            fftproduct(U8,G,U8G,row,col);
            fftproduct(U9,G,U9G,row,col);
            
            // inverting the fft's product results
            fftw_execute(pinvU1);
            iu1 *= 1.0/(col*row);
            fftw_execute(pinvU2);
            iu2 *= 1.0/(col*row);
            fftw_execute(pinvU3);
            iu3 *= 1.0/(col*row);
            fftw_execute(pinvU4);
            iu4 *= 1.0/(col*row);
            fftw_execute(pinvU5);
            iu5 *= 1.0/(col*row);
            fftw_execute(pinvU6);
            iu6 *= 1.0/(col*row);
            fftw_execute(pinvU7);
            iu7 *= 1.0/(col*row);
            fftw_execute(pinvU8);
            iu8 *= 1.0/(col*row);
            fftw_execute(pinvU9);
            iu9 *= 1.0/(col*row);
            
            /* -----------------------------------*/
            /* Polynomial                         */
            /* -----------------------------------*/
            for (int i = 0; i < length; i++)
            {
                // applying first part of the polynomial approximation
                transform_polynom_coef(poly,p,u1.data[i],u2.data[i],
                                       u3.data[i],u4.data[i],u5.data[i],
                                       u6.data[i],u7.data[i],u8.data[i],
                                       u9.data[i]);
                
                // applyiing second part of the polynomial approximation
                R.data[i] = - p[0]
                -p[1] * iu1.data[i]
                -p[2] * iu2.data[i]
                -p[3] * iu3.data[i]
                -p[4] * iu4.data[i]
                -p[5] * iu5.data[i]
                -p[6] * iu6.data[i]
                -p[7] * iu7.data[i]
                -p[8] * iu8.data[i]
                -p[9] * iu9.data[i];
                
                // security patch to ensure convergence of the process
                if(R.data[i] > 1.0)
                    R.data[i] = 1.0;
                else
                    if(R.data[i] < -1.0)
                        R.data[i] = -1.0;
                
            }
            
            // normalizing R term to estandarize results
            R *= (1.0/R.maxabsval());
            
            // saving Ik to evaluate Ik+1
            Im_k = Im_kp1;
            
            /*
             applying main equation:
             Ik+1 = (Ik + dt(alpha*1/2 + beta*I0 + R))/(1 + dt*(alpha + beta))
             */

            for (int x = 0; x < length; x++){
                Im_kp1.data[x] =(Im_k.data[x] +
                                 dt*(alpha*mu + beta*Im0[color].data[x]
                                     + 0.5*R.data[x] )
                                 )*norm;
            }
            
            
            // Reached convergence?
            // evaluation difference Ik+1 - Ik
            // to decide whether the process should end or not.
            difference = Im_k.Dist(Im_kp1);
            
            // if verbose write some process information
            if(verbose)
            {
                fprintf(stderr,"diff=%lf",difference);
                fprintf(stderr,"it=%d,",iteration);
                fprintf(stderr,"range:(%lf,%lf)",Im_kp1.minval(),Im_kp1.maxval());
                fprintf(stderr,"mean value=%lf\n",Im_kp1.meanvalue());
                // something is wrong with the process?
                if(Im_kp1.maxabsval() > 1.0)
                {
                    fprintf(stderr,"\nwarning: image has");
                    fprintf(stderr,"invalid values > 1%lf\n",Im_kp1.maxabsval());
                }
            }
            
            iteration++; /// next iteration
        }
        
        //  rescaled as a final step
        Im_kp1.rescale(0.0,1.0);
        
        // if verbose write some process information
        if(verbose)
        {
            fprintf(stderr,"\n");
            fprintf(stderr,"[%d]original range:(%lf,%lf),",
                    color,Im0[color].minval(), Im0[color].maxval());
            
            fprintf(stderr,"original mean value=%lf\n",
                    Im0[color].meanvalue());
            
            fprintf(stderr,"[%d]new range:(%lf,%lf),",
                    color,Im_kp1.minval(), Im_kp1.maxval());

            fprintf(stderr,"original mean value=%lf\n",
                    Im_kp1.meanvalue());
            
        }
        
        Im0[color] = Im_kp1;
        iteration=0;
        difference=1000.0;
    }

    // Free data
    fftw_free(U1);
    fftw_free(U2);
    fftw_free(U3);
    fftw_free(U4);
    fftw_free(U5);
    fftw_free(U6);
    fftw_free(U7);
    fftw_free(U8);
    fftw_free(U9);
    
    fftw_free(U1G);
    fftw_free(U2G);
    fftw_free(U3G);
    fftw_free(U4G);
    fftw_free(U5G);
    fftw_free(U6G);
    fftw_free(U7G);
    fftw_free(U8G);
    fftw_free(U9G);
    
    fftw_destroy_plan(pinvU1);
    fftw_destroy_plan(pinvU2);
    fftw_destroy_plan(pinvU3);
    fftw_destroy_plan(pinvU4);
    fftw_destroy_plan(pinvU5);
    fftw_destroy_plan(pinvU6);
    fftw_destroy_plan(pinvU7);
    fftw_destroy_plan(pinvU8);
    fftw_destroy_plan(pinvU9);
    
    fftw_destroy_plan(pU1);
    fftw_destroy_plan(pU2);
    fftw_destroy_plan(pU3);
    fftw_destroy_plan(pU4);
    fftw_destroy_plan(pU5);
    fftw_destroy_plan(pU6);
    fftw_destroy_plan(pU7);
    fftw_destroy_plan(pU8);
    fftw_destroy_plan(pU9);
    
    fftw_free(G);
    fftw_destroy_plan(pG);
    free(p);
    
   
}
