 /*
 This program is free software: you can use, modify and/or
 redistribute it under the terms of the simplified BSD
 License. You should have received a copy of this license along
 this program. If not, see
 <http://www.opensource.org/licenses/bsd-license.html>.
 
  */

#include <inout.h>
#include <Image.h>
#include <contrast_enhancement.h>
#include <omp.h>
/**
 * @file    main_contrast_enhancement_cmd.cpp
 * @brief   Main file for contrast_enhancement.
 *
 * This file deals with I/O and calls the local 
 * contrast enhacement algorithm with an RGB color image.
 * It defines the main function called from the command-line
 * with some arguments.
 *
 *
 * @author Sira Ferradans and Rodrigo Palma-Amestoy
 */

/** @brief Function that displays how to correctly use the program */
void how_to_use()
{
    fprintf(stderr,"Use: contrast_enhancement input output");
    fprintf(stderr,"contrast_function alpha kernel_type kernel_size [verbose]\n");
    fprintf(stderr,"\t-input: input tiff image\n");
    fprintf(stderr,"\t-output: output tiff image\n");
    fprintf(stderr,"\t-contrast_function: contrast");
    fprintf(stderr,"function to be used\n\t\t");
    fprintf(stderr,"(1) Entropic ACE (phi=log)\n\t\t");
    fprintf(stderr,"(2) Entropic Michelson (phi=-M)\n\t\t");
    fprintf(stderr,"(3) Entropic MIN-MAX (phi=id)\n");
    fprintf(stderr,"\t-alpha: attachment to grey value");
    fprintf(stderr,"       (range: 1 - inf, typical: 1.1)\n");
    fprintf(stderr,"\t-beta: attachment to original image");
    fprintf(stderr,"        (range: 0 - inf, typical: 1)\n");
    fprintf(stderr,"\t-kernel_type: \n\t\t");
    fprintf(stderr,"       (1) linear kernel\n\t\t(other) Gaussian kernel\n");
    fprintf(stderr,"\t-kernel_size: kernel's parameter");
    fprintf(stderr,"       (range: 0.01 - 1.0, typical: 0.1)\n");
    fprintf(stderr,"\t-verbose: enable verbose mode");
    fprintf(stderr,"(1), default not verbose\n");
    exit(1);
};

/**
 * @brief   Main function.
 *
 * The main function is used to:
 * @li get, and check command-line parameters.
 * @li load the image and polynom.
 * @li launch the contrast enhancement process.
 */
int main(int argc,char **argv)
{
 
    // check if the input is correct
    if(argc < 8)
        how_to_use();
        
 
    // reading input params
    argv++;
    char * file_in=*argv++;
    char * file_out=*argv++;
    int cont_function = atoi(*argv++);
    double alpha = atof(*argv++);
    double beta = atof(*argv++);
    int type_kernel=atoi(*argv++);
    double sigma = atof(*argv++);
    int verbose = 0;
    if(argc > 8)
        verbose = atoi(*argv++);
    
    // reading polynomial coefficient of contrast functions estimation
    FILE *file;
    char poly_name[100];
    
    if(cont_function == 1)
        strcpy (poly_name,"polynomials/ACE_epsilon20.dat");
    else if(cont_function == 2)
        strcpy (poly_name,"polynomials/michelson_epsilon20.dat");
    else if(cont_function == 3)
        strcpy (poly_name,"polynomials/minmax_epsilon20.dat");
   
    double poly[55];
    char polyaux[100];
    
    file = fopen(poly_name,"r");
    for (int i = 0; i < 55; i++)
    {
        fscanf (file, "%s\n", polyaux);
        poly[i] = strtod(polyaux, NULL);
    }
    fclose (file);
    
    // reading original image to be enhanced
    Image I0[3];
    read(I0,file_in);
    
    
    // timing the algorithm
    int c1,cInit=omp_get_wtime();
   
    // executing color enhancement algorithm, modification of I0
    contrast_enhancement(I0,poly,type_kernel, sigma, alpha, beta, verbose);
    
    // timing the last step of the algorithm
    c1=omp_get_wtime();
    fprintf(stderr,"\nComplete execution done in:");
    fprintf(stderr,"%f \n\n", (double)(c1 - cInit));
 
    write(file_out,I0); // writing image to the disc
}

